{************************************************************************************************}
{* 4-bit full-adder-subtractor (4-bit incrementer-decrementer) unit for LogicGates8 application *}
{************************************************************************************************}

unit addsub4f;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls, StdCtrls, Grids, common;

type
  {************}
  { TfAddSub4f }
  {************}
  TfAddSub4f = class(TForm)
    Shape1: TShape;
    Shape2: TShape;
    StaticText1: TStaticText;
    Label2, Label8, Label9, Label10, Label18, Label3, Label4, Label11, laOperation: TLabel;
    laA0, laA1, laA2, laA3, laB0, laB1, laB2, laB3: TLabel;
    laS0, laS1, laS2, laS3, laM: TLabel;
    Shape20, Shape21, Shape22, Shape23, Shape28, Shape25, Shape32, Shape36, Shape38: TShape;
    Shape58, Shape63, Shape78, Shape87, Shape94, Shape98, Shape99: TShape;
    Shape45, Shape50, Shape51, Shape52, Shape53, Shape54, Shape55: TShape;
    Shape56, Shape57, Shape59, Shape60, Shape61, Shape62, Shape64, Shape65: TShape;
    Shape66, Shape67, Shape68, Shape69, Shape70, Shape71, Shape72, Shape73, Shape74, Shape75: TShape;
    Shape76, Shape77, Shape79, Shape80, Shape81, Shape82, Shape83, Shape85: TShape;
    Shape86, Shape88, Shape89, Shape90, Shape91, Shape92, Shape93, Shape95: TShape;
    Shape96, Shape97, Shape100, Shape101, Shape102, Shape103, Shape104, Shape105: TShape;
    Shape106, Shape107, Shape108, Shape109, Shape110, Shape111, Shape112, Shape113: TShape;
    Shape114, Shape115, Shape154, Shape155, Shape156, Shape157, Shape158: TShape;
    Shape116, Shape117, Shape118, Shape119, Shape120, Shape121, Shape122, Shape123: TShape;
    Shape124, Shape125, Shape126, Shape127, Shape128, Shape129, Shape130, Shape131, Shape132: TShape;
    Shape133, Shape134, Shape135, Shape136, Shape137, Shape138, Shape139, Shape140, Shape141: TShape;
    Shape142, Shape143, Shape144, Shape145, Shape146, Shape147, Shape148, Shape149, Shape150, Shape151, Shape152: TShape;
    Image1, Image2, Image3, Image4, Image5, Image6, Image7, Image8, Image9: TImage;
    imSwitchA0, imSwitchA1, imSwitchA2, imSwitchA3, imSwitchM: TImage;
    shSwitchA0a, shSwitchA0b, shSwitchA1a, shSwitchA1b: TShape;
    shSwitchA2a, shSwitchA2b, shSwitchA3a, shSwitchA3b, shSwitchMa, shSwitchMb: TShape;
    shLedA0, shLedA1, shLedA2, shLedA3: TShape;
    shLedB0, shLedB1, shLedB2, shLedB3: TShape;
    shLedS0, shLedS1, shLedS2, shLedS3: TShape;
    shLedCout, shLedM: TShape;
    sgSub: TStringGrid;
    btClose: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure imSwitchA0Click(Sender: TObject);
    procedure imSwitchA1Click(Sender: TObject);
    procedure imSwitchA2Click(Sender: TObject);
    procedure imSwitchA3Click(Sender: TObject);
    procedure imSwitchMClick(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
  private
    aLeds: TLeds;
    aSwitches: TSwitches;
    aInput, aOutput: TStatus;
  end;

const
  // Switches description: Data = objects positions (transparent image, long bar shape, short bar shape)
  DSwitches: array[0..8] of TSwitchData = (
    (Name: 'SwitchA0';
     Data: (512, 534, 514)),
    (Name: 'SwitchA1';
     Data: (378, 400, 380)),
    (Name: 'SwitchA2';
     Data: (246, 268, 248)),
    (Name: 'SwitchA3';
     Data: (112, 134, 114)),
    (Name: 'undef';
     Data: (-1, -1, -1)),
    (Name: 'undef';
     Data: (-1, -1, -1)),
    (Name: 'undef';
     Data: (-1, -1, -1)),
    (Name: 'undef';
     Data: (-1, -1, -1)),
    (Name: 'SwitchM';
     Data: (636, 658, 638))
  );

var
  fAddSub4f: TfAddSub4f;

implementation

{$R *.lfm}

{ 4-bit-full-adder-subtractor: Set switch position; do binary addition resp. subtraction; turn LEDs on/off; fill-in stringgrid }

procedure SwitchAndAddSub(DSwitches: array of TSwitchData; Switches: TSwitches; var InputStatus: TStatus;
  Leds: TLeds; var OutputStatus: TStatus; IX: Integer);

var
  Exp, I: Integer;
  A, B, C, Sout, Cout, S1, S2, C1, C2: Boolean;
  Dec: array[0..3] of Integer;
  Bin, Compl: TBinary;

begin
  // Change current input from 0 to Vs (and vice-versa)
  InputStatus[IX] := not InputStatus[IX];
  // Corresponding LED status
  if IX = 8 then
    // Input 8 = control signal (LED 13)
    OutputStatus[13] := InputStatus[8]
  else
    // Input 0 - 7 = input bits of A and B (LEDs 0 - 7)
    OutputStatus[IX] := InputStatus[IX];
  // Set switch positions
  SwitchesOnOff(DSwitches, Switches, InputStatus);
  // Adapt grid label
  if InputStatus[8] then
    fAddSub4f.laOperation.Caption := '4 bits binary SUBTRACTION:'
  else
    fAddSub4f.laOperation.Caption := '4 bits binary ADDITION:';
  // Do the addition, using 4 1-bit-full-adders:
  // Addition of A and B (for M = 0); addition of A and the 2's complement of B (for M = 1)
  // The code logic is identical to what happens in the electronics circuit:
  // Cin of the first adder depends on M (for subtraction, Cin = 1: this allows to add the bit needed for 2's complement)
  // The B inputs are inverted or not, depending on the result of the corr. XOR
  Cout := OutputStatus[13];                                                    // Cin of first adder set to 0 or 1 (depending on control M)
  for I := 0 to 3 do begin
    A := OutputStatus[I];                                                      // actual bit for A input
    B := OutputStatus[13] xor OutputStatus[I + 4];                             // actual bit for B input, inverted or not (depending on control M)
    C := Cout;                                                                 // Cin = previous Cout
    // 1-bit-full-adder binary addition
    Adder(A, B, C, S1, C1, S2, C2, Sout, Cout);
    // Actual bit sum LED status
    OutputStatus[I + 8] := SOut;
  end;
  // Cout of the last adder is carry value of 4-bit-addition, resp. indicates polarity of 4-bit-subtraction result
  OutputStatus[12] := Cout;
  // Turn all LEDs on/off
  LedsOnOff(Leds, OutputStatus);
  // Fill binary values into stringgrid
  for I := 0 to 3 do begin
    fAddSub4f.sgSub.Cells[5 - I, 1] := ' ' + IntToStr(Bits[OutputStatus[I]]);
    fAddSub4f.sgSub.Cells[5 - I, 2] := ' ' + IntToStr(Bits[OutputStatus[I + 4]]);
    fAddSub4f.sgSub.Cells[5 - I, 3] := ' ' + IntToStr(Bits[OutputStatus[I + 8]]);
    Bin[3 - I] := OutputStatus[I + 8];
  end;
  // For addition, fill in carry (if is 1)
  fAddSub4f.sgSub.Cells[1, 3] := '';
  if not OutputStatus[13] then begin
    // Addition
    if OutputStatus[12] then
      // There actually is a carry
      fAddSub4f.sgSub.Cells[1, 3] := ' ' + IntToStr(Bits[OutputStatus[12]]);
  end;
  // Calculate and fill in result's 2nd complement
  Compl := Complement2(Bin);
  for I := 0 to 3 do begin
    fAddSub4f.sgSub.Cells[I + 2, 4] := '';
    // Display 2's complement only for subtraction
    if OutputStatus[13] then
      fAddSub4f.sgSub.Cells[I + 2, 4] := ' ' + IntToStr(Bits[Compl[I]]);
  end;
  // Calculate decimal values
  Dec[0] := 0; Dec[1] := 0; Dec[2] := 0; Exp := 1;
  for I := 0 to 3 do begin
    Dec[0] += Bits[OutputStatus[I]] * Exp; Dec[1] += Bits[OutputStatus[I + 4]] * Exp; Dec[2] += Bits[OutputStatus[I + 8]] * Exp;
    Exp *= 2;
  end;
  // Adjust decimal value of result
  if OutputStatus[13] then begin
    // Subtraction: Decimal result may be positive or negative
    if Dec[0] < Dec[1] then
      Dec[2] := -(16 - Dec[2]);
  end
  else begin
    // Addition: Consideration of last adder's Cout to get correct decimal result
    Dec[2] += Bits[OutputStatus[12]] * 16;
  end;
  // Fill-in decimal values
  for I := 0 to 2 do begin
    fAddSub4f.sgSub.Cells[6, I + 1] := ' ' + IntToStr(Dec[I]);
    if Abs(Dec[I]) < 10 then
      fAddSub4f.sgSub.Cells[6, I + 1] := ' ' + fAddSub4f.sgSub.Cells[6, I + 1];
    if Dec[I] >= 0 then
      fAddSub4f.sgSub.Cells[6, I + 1] := ' ' + fAddSub4f.sgSub.Cells[6, I + 1];
  end;
end;

{************}
{ TfAddSub4f }
{************}

{ Application start: Initialisation }

procedure TfAddSub4f.FormCreate(Sender: TObject);

begin
  // Arrays with switches (transparent image and bar shapes) and their "on/off" status
  SetLength(aSwitches, 9); SetLength(aInput, 9);
  aSwitches[0].Switch0 := imSwitchA0; aSwitches[0].Switch1 := shSwitchA0a; aSwitches[0].Switch2 := shSwitchA0b;
  aSwitches[1].Switch0 := imSwitchA1; aSwitches[1].Switch1 := shSwitchA1a; aSwitches[1].Switch2 := shSwitchA1b;
  aSwitches[2].Switch0 := imSwitchA2; aSwitches[2].Switch1 := shSwitchA2a; aSwitches[2].Switch2 := shSwitchA2b;
  aSwitches[3].Switch0 := imSwitchA3; aSwitches[3].Switch1 := shSwitchA3a; aSwitches[3].Switch2 := shSwitchA3b;
  aSwitches[4].Switch0 := nil; aSwitches[4].Switch1 := nil; aSwitches[4].Switch2 := nil;
  aSwitches[5].Switch0 := nil; aSwitches[5].Switch1 := nil; aSwitches[5].Switch2 := nil;
  aSwitches[6].Switch0 := nil; aSwitches[6].Switch1 := nil; aSwitches[6].Switch2 := nil;
  aSwitches[7].Switch0 := nil; aSwitches[7].Switch1 := nil; aSwitches[7].Switch2 := nil;
  aSwitches[8].Switch0 := imSwitchM;  aSwitches[8].Switch1 := shSwitchMa;  aSwitches[8].Switch2 := shSwitchMb;
  // Arrays with LEDs (shape) and their "on/off" status
  SetLength(aLeds, 14); SetLength(aOutput, 14);
  aLeds[0] := shLedA0; aLeds[1] := shLedA1; aLeds[2] := shLedA2; aLeds[3] := shLedA3;
  aLeds[4] := shLedB0; aLeds[5] := shLedB1; aLeds[6] := shLedB2; aLeds[7] := shLedB3;
  aLeds[8] := shLedS0; aLeds[9] := shLedS1; aLeds[10] := shLedS2; aLeds[11] := shLedS3;
  aLeds[12] := shLedCout; aLeds[13] := shLedM;
  // Use subscripts in input/output labels
  laA0.Caption := 'A' + SUB_Digits[0]; laA1.Caption := 'A' + SUB_Digits[1];
  laA2.Caption := 'A' + SUB_Digits[2]; laA3.Caption := 'A' + SUB_Digits[3];
  laB0.Caption := 'B' + SUB_Digits[0]; laB1.Caption := 'B' + SUB_Digits[1];
  laB2.Caption := 'B' + SUB_Digits[2]; laB3.Caption := 'B' + SUB_Digits[3];
  laS0.Caption := 'S' + SUB_Digits[0]; laS1.Caption := 'S' + SUB_Digits[1];
  laS2.Caption := 'S' + SUB_Digits[2]; laS3.Caption := 'S' + SUB_Digits[3];
end;

{ Form show-up: Initialize 4-bit incrementer-decrementer simulation }

procedure TfAddSub4f.FormActivate(Sender: TObject);

var
  I: Integer;

begin
  // All inputs for A, the number to be incremented/decremented, to logic 0
  // For B, the increment/decrement, least significant bit to logic 1 (increment/decrement = addition/subtraction of 0001)
  // Control switch M = off (logic 0): Addition
  for I := 0 to Length(aInput) - 1 do begin
    if I = 4 then
      aInput[I] := True
    else
      aInput[I] := False;
  end;
  SwitchesOnOff(DSwitches, aSwitches, aInput);
  // All outputs to logic 0, except B0 (least significant bit of increment) and S0 (0 + 1 = 1)
  for I := 0 to Length(aOutput) - 1 do begin
    if (I = 4) or (I = 8) then
      aOutput[I] := True
    else
      aOutput[I] := False;
  end;
  LedsOnOff(aLeds, aOutput);
end;

{ Switch A0, A1, A2, A3 activated: Do 4-bit incrementer-decrementer simulation for actual switch positions }

procedure TfAddSub4f.imSwitchA0Click(Sender: TObject);

begin
  SwitchAndAddSub(DSwitches, aSwitches, aInput, aLeds, aOutput, 0);
end;

procedure TfAddSub4f.imSwitchA1Click(Sender: TObject);

begin
  SwitchAndAddSub(DSwitches, aSwitches, aInput, aLeds, aOutput, 1);
end;

procedure TfAddSub4f.imSwitchA2Click(Sender: TObject);

begin
  SwitchAndAddSub(DSwitches, aSwitches, aInput, aLeds, aOutput, 2);
end;

procedure TfAddSub4f.imSwitchA3Click(Sender: TObject);

begin
  SwitchAndAddSub(DSwitches, aSwitches, aInput, aLeds, aOutput, 3);
end;

{ Switch "M" activated: Update 4-bit-full-adder-subtractor simulation for actual switch positions }

procedure TfAddSub4f.imSwitchMClick(Sender: TObject);

begin
  SwitchAndAddSub(DSwitches, aSwitches, aInput, aLeds, aOutput, 8);
end;

{ Button "Close": Close the window }

procedure TfAddSub4f.btCloseClick(Sender: TObject);

begin
  Close;
end;

end.

